import json
import logging
import uuid
import os

from unittest.mock import Mock
from pathlib import Path

import pytest
import nbformat

from jinja2 import Environment
from notebook.services.contents.filemanager import FileContentsManager
from tornado.testing import AsyncTestCase, gen_test
from tornado.web import Application, HTTPError
from tornado.httpserver import HTTPRequest
from traitlets.config import Config

from bookstore.bookstore_config import BookstoreSettings
from bookstore.clone import (
    build_notebook_model,
    build_file_model,
    BookstoreCloneHandler,
    BookstoreCloneAPIHandler,
    validate_relpath,
    BookstoreFSCloneHandler,
    BookstoreFSCloneAPIHandler,
)
from bookstore.utils import TemporaryWorkingDirectory

from . import test_dir


log = logging.getLogger('test_clone')


def test_build_notebook_model():
    content = nbformat.v4.new_notebook()
    expected = {
        "type": "notebook",
        "format": "json",
        "content": content,
        "name": "my_notebook_name.ipynb",
        "path": "test_directory/my_notebook_name.ipynb",
    }
    path = "./test_directory/my_notebook_name.ipynb"
    nb_content = nbformat.writes(content)
    assert build_notebook_model(nb_content, path) == expected


def test_build_file_model():
    content = "my fancy file"
    expected = {
        "type": "file",
        "format": "text",
        "content": content,
        "name": "file_name.txt",
        "path": "test_directory/file_name.txt",
    }
    path = "./test_directory/file_name.txt"
    assert build_file_model(content, path) == expected


class TestCloneHandler(AsyncTestCase):
    def setUp(self):
        super().setUp()
        self.mock_application = Mock(
            spec=Application, ui_methods={}, ui_modules={}, settings={'jinja2_env': Environment()}
        )

    def get_handler(self, uri, app=None):
        if app is None:
            app = self.mock_application
        connection = Mock(context=Mock(protocol="https"))
        payload_request = HTTPRequest(
            method='GET',
            uri=uri,
            headers={"Host": "localhost:8888"},
            body=None,
            connection=connection,
        )
        return BookstoreCloneHandler(app, payload_request)

    @gen_test
    async def test_get_no_param(self):
        empty_handler = self.get_handler('/bookstore/clone')
        with pytest.raises(HTTPError):
            await empty_handler.get()

    @gen_test
    async def test_get_no_bucket(self):
        no_bucket_handler = self.get_handler('/bookstore/clone?s3_bucket=&s3_key=hi')
        with pytest.raises(HTTPError):
            await no_bucket_handler.get()

    @gen_test
    async def test_get_no_object_key(self):
        no_object_key_handler = self.get_handler('/bookstore/clone?s3_bucket=hello&s3_key=')
        with pytest.raises(HTTPError):
            await no_object_key_handler.get()

    @gen_test
    async def test_get_success(self):
        success_handler = self.get_handler('/bookstore/clone?s3_bucket=hello&s3_key=my_key')
        await success_handler.get()

    def test_gen_template_params(self):
        success_handler = self.get_handler('/bookstore/clone?s3_bucket=hello&s3_key=my_key')
        expected = {
            'post_model': {'s3_bucket': 'hello', 's3_key': 'my_key'},
            'clone_api_url': 'https://localhost:8888/api/bookstore/clone',
            'redirect_contents_url': 'https://localhost:8888',
            'source_description': 'my_key from hello',
        }
        success_handler = self.get_handler('/bookstore/clone?s3_bucket=hello&s3_key=my_key')
        output = success_handler.construct_template_params(
            s3_bucket="hello", s3_object_key="my_key"
        )
        assert expected == output

    def test_gen_template_params_base_url(self):
        base_url_list = ['/my_base_url', '/my_base_url/', 'my_base_url/', 'my_base_url']
        expected = {
            'post_model': {'s3_bucket': 'hello', 's3_key': 'my_key'},
            'clone_api_url': 'https://localhost:8888/my_base_url/api/bookstore/clone',
            'redirect_contents_url': 'https://localhost:8888',
            'source_description': 'my_key from hello',
        }
        for base_url in base_url_list:
            mock_app = Mock(
                spec=Application,
                ui_methods={},
                ui_modules={},
                settings={'jinja2_env': Environment(), "base_url": base_url},
            )

            success_handler = self.get_handler(
                '/bookstore/clone?s3_bucket=hello&s3_key=my_key', app=mock_app
            )
            output = success_handler.construct_template_params(
                s3_bucket="hello", s3_object_key="my_key"
            )
            assert expected == output


class TestCloneAPIHandler(AsyncTestCase):
    def setUp(self):
        super().setUp()
        mock_settings = {
            "BookstoreSettings": {
                "s3_access_key_id": "mock_id",
                "s3_secret_access_key": "mock_access",
            }
        }
        config = Config(mock_settings)

        self.mock_application = Mock(
            spec=Application,
            ui_methods={},
            ui_modules={},
            settings={
                'jinja2_env': Environment(),
                "config": config,
                "contents_manager": FileContentsManager(),
            },
        )

    def post_handler(self, body_dict, app=None):
        if app is None:
            app = self.mock_application
        body = json.dumps(body_dict).encode('utf-8')
        payload_request = HTTPRequest(
            method='POST', uri="/api/bookstore/clone", headers=None, body=body, connection=Mock()
        )
        return BookstoreCloneAPIHandler(app, payload_request)

    @gen_test
    async def test_post_no_body(self):
        post_body_dict = {}
        empty_handler = self.post_handler(post_body_dict)
        with pytest.raises(HTTPError):
            await empty_handler.post()

    @gen_test
    async def test_post_empty_bucket(self):
        post_body_dict = {"s3_key": "my_key", "s3_bucket": ""}
        empty_bucket_handler = self.post_handler(post_body_dict)
        with pytest.raises(HTTPError):
            await empty_bucket_handler.post()

    @gen_test
    async def test_post_empty_key(self):
        post_body_dict = {"s3_key": "", "s3_bucket": "my_bucket"}
        empty_key_handler = self.post_handler(post_body_dict)
        with pytest.raises(HTTPError):
            await empty_key_handler.post()

    @gen_test
    async def test_post_nonsense_params(self):
        post_body_dict = {"s3_key": "my_key", "s3_bucket": "my_bucket"}
        success_handler = self.post_handler(post_body_dict)
        with pytest.raises(HTTPError):
            await success_handler.post()

    @gen_test
    async def test_private_clone_nonsense_params(self):
        s3_bucket = "my_key"
        s3_object_key = "my_bucket"
        post_body_dict = {"s3_key": s3_object_key, "s3_bucket": s3_bucket}
        success_handler = self.post_handler(post_body_dict)
        with pytest.raises(HTTPError):
            await success_handler._clone(s3_bucket, s3_object_key)

    def test_build_post_response_model(self):
        content = "some arbitrary content"
        expected = {
            "type": "file",
            "format": "text",
            "content": content,
            "name": "file_name.txt",
            "path": "test_directory/file_name.txt",
            "s3_path": "s3://my_bucket/original_key/may_be_different_than_storage.txt",
            'versionID': "eeee222eee",
        }

        s3_bucket = "my_bucket"
        s3_object_key = "original_key/may_be_different_than_storage.txt"
        obj = {'VersionId': "eeee222eee"}
        model = {
            "type": "file",
            "format": "text",
            "content": content,
            "name": "file_name.txt",
            "path": "test_directory/file_name.txt",
        }
        handler = self.post_handler({})
        actual = handler.build_post_response_model(model, obj, s3_bucket, s3_object_key)
        assert actual == expected

    @gen_test
    async def test_build_text_content_model(self):
        content = "some content"
        expected = {
            "type": "file",
            "format": "text",
            "content": content,
            "name": "file_name.txt",
            "path": "test_directory/file_name.txt",
        }

        path = "test_directory/file_name.txt"
        success_handler = self.post_handler({})
        model = success_handler.build_content_model(content, path)
        assert model == expected

    @gen_test
    async def test_build_notebook_content_model(self):
        content = nbformat.v4.new_notebook()
        expected = {
            "type": "notebook",
            "format": "json",
            "content": content,
            "name": "file_name.ipynb",
            "path": "test_directory/file_name.ipynb",
        }

        str_content = nbformat.writes(content)

        path = "test_directory/file_name.ipynb"
        success_handler = self.post_handler({})
        model = success_handler.build_content_model(str_content, path)
        assert model == expected


def test_validate_relpath():
    relpath = 'hi'
    settings = BookstoreSettings(fs_cloning_basedir="/anything")
    fs_clonepath = validate_relpath(relpath, settings, log)
    assert fs_clonepath == Path("/anything/hi")


def test_validate_relpath_empty_relpath(caplog):
    relpath = ''
    settings = BookstoreSettings(fs_cloning_basedir="/anything")
    with pytest.raises(HTTPError):
        with caplog.at_level(logging.INFO):
            fs_clonepath = validate_relpath(relpath, settings, log)
    assert "Request received with empty relpath." in caplog.text


def test_validate_relpath_escape_basedir(caplog):
    relpath = '../hi'
    settings = BookstoreSettings(fs_cloning_basedir="/anything")
    with pytest.raises(HTTPError):
        with caplog.at_level(logging.INFO):
            fs_clonepath = validate_relpath(relpath, settings, log)
    assert f"Request to clone from a path outside of base directory" in caplog.text


class TestFSCloneHandler(AsyncTestCase):
    def setUp(self):
        super().setUp()
        mock_settings = {"BookstoreSettings": {"fs_cloning_basedir": test_dir}}
        self.config = Config(mock_settings)
        self.mock_application = Mock(
            spec=Application,
            ui_methods={},
            ui_modules={},
            settings={'jinja2_env': Environment(), 'config': self.config},
        )

    def get_handler(self, uri, app=None):
        if app is None:
            app = self.mock_application
        connection = Mock(context=Mock(protocol="https"))
        payload_request = HTTPRequest(
            method='GET',
            uri=uri,
            headers={"Host": "localhost:8888"},
            body=None,
            connection=connection,
        )
        return BookstoreFSCloneHandler(app, payload_request)

    @gen_test
    async def test_get_no_param(self):
        empty_handler = self.get_handler('/bookstore/fs-clone')
        with pytest.raises(HTTPError):
            await empty_handler.get()

    @gen_test
    async def test_get_empty_relpath(self):
        empty_relpath_handler = self.get_handler('/bookstore/fs-clone?relpath=')
        with pytest.raises(HTTPError):
            await empty_relpath_handler.get()

    @gen_test
    async def test_get_escape_basedir(self):
        escape_basedir_handler = self.get_handler('/bookstore/fs-clone?relpath=../hi')
        with pytest.raises(HTTPError):
            await escape_basedir_handler.get()

    @gen_test
    async def test_get_success(self):
        success_handler = self.get_handler('/bookstore/fs-clone?relpath=my/test/path.ipynb')
        await success_handler.get()

    def test_gen_template_params(self):
        expected = {
            'post_model': {'relpath': 'my/test/path.ipynb'},
            'clone_api_url': 'https://localhost:8888/api/bookstore/fs-clone',
            'redirect_contents_url': 'https://localhost:8888',
            'source_description': '/Users/jupyter/my/test/path.ipynb',
        }
        success_handler = self.get_handler('/bookstore/fs-clone?relpath=my/test/path.ipynb')
        output = success_handler.construct_template_params(
            relpath='my/test/path.ipynb', fs_clonepath='/Users/jupyter/my/test/path.ipynb'
        )
        assert expected == output

    def test_gen_template_params_base_url(self):
        base_url_list = ['/my_base_url', '/my_base_url/', 'my_base_url/', 'my_base_url']
        expected = {
            'post_model': {'relpath': 'my/test/path.ipynb'},
            'clone_api_url': 'https://localhost:8888/my_base_url/api/bookstore/fs-clone',
            'redirect_contents_url': 'https://localhost:8888',
            'source_description': '/Users/jupyter/my/test/path.ipynb',
        }
        for base_url in base_url_list:
            mock_app = Mock(
                spec=Application,
                ui_methods={},
                ui_modules={},
                settings={'jinja2_env': Environment(), "base_url": base_url, "config": self.config},
            )

            success_handler = self.get_handler(
                '/bookstore/fs-clone?relpath=my/test/path.ipynb', app=mock_app
            )
            output = success_handler.construct_template_params(
                relpath='my/test/path.ipynb', fs_clonepath='/Users/jupyter/my/test/path.ipynb'
            )
            assert expected == output


class TestFSCloneAPIHandler(AsyncTestCase):
    def setUp(self):
        super().setUp()
        mock_settings = {
            "BookstoreSettings": {"fs_cloning_basedir": os.path.join(test_dir, 'test_files')}
        }
        config = Config(mock_settings)

        self.mock_application = Mock(
            spec=Application,
            ui_methods={},
            ui_modules={},
            settings={
                'jinja2_env': Environment(),
                "config": config,
                "contents_manager": FileContentsManager(),
            },
        )

    def post_handler(self, body_dict, app=None):
        if app is None:
            app = self.mock_application
        body = json.dumps(body_dict).encode('utf-8')
        payload_request = HTTPRequest(
            method='POST', uri="/api/bookstore/fs-clone", headers=None, body=body, connection=Mock()
        )
        return BookstoreFSCloneAPIHandler(app, payload_request)

    @gen_test
    async def test_post_no_body(self):
        post_body_dict = {}
        empty_handler = self.post_handler(post_body_dict)
        with pytest.raises(HTTPError):
            await empty_handler.post()

    @gen_test
    async def test_post_empty_relpath(self):
        post_body_dict = {"relpath": ""}
        empty_relpath_handler = self.post_handler(post_body_dict)
        with pytest.raises(HTTPError):
            await empty_relpath_handler.post()

    @gen_test
    async def test_post_basedir_escape(self):
        post_body_dict = {"relpath": "../myfile.txt"}
        empty_relpath_handler = self.post_handler(post_body_dict)
        with pytest.raises(HTTPError):
            await empty_relpath_handler.post()

    @gen_test
    async def test_post_nonsense_params(self):
        post_body_dict = {"relpath": str(uuid.uuid4())}
        success_handler = self.post_handler(post_body_dict)
        with pytest.raises(HTTPError):
            await success_handler.post()

    @gen_test
    async def test_post_success_notebook(self):
        post_body_dict = {"relpath": 'EmptyNotebook.ipynb'}
        with open(os.path.join(test_dir, 'test_files/EmptyNotebook.ipynb'), 'r') as f:
            expected = json.load(f)
        success_handler = self.post_handler(post_body_dict)
        setattr(success_handler, '_transforms', [])

        with TemporaryWorkingDirectory() as tmp:
            await success_handler.post()
            with open('EmptyNotebook.ipynb') as f:
                actual = json.load(f)
        assert actual == expected

    @gen_test
    async def test_build_text_content_model(self):
        content = "some content"
        expected = {
            "type": "file",
            "format": "text",
            "content": content,
            "name": "file_name.txt",
            "path": "test_directory/file_name.txt",
        }

        path = "test_directory/file_name.txt"
        post_handler = self.post_handler({})
        model = post_handler.build_content_model(content, path)
        assert model == expected

    @gen_test
    async def test_build_notebook_content_model(self):
        content = nbformat.v4.new_notebook()
        expected = {
            "type": "notebook",
            "format": "json",
            "content": content,
            "name": "file_name.ipynb",
            "path": "test_directory/file_name.ipynb",
        }

        str_content = nbformat.writes(content)

        path = "test_directory/file_name.ipynb"
        post_handler = self.post_handler({})
        model = post_handler.build_content_model(str_content, path)
        assert model == expected



"""Configuration settings for bookstore."""
import logging
from pathlib import Path

from traitlets import Integer, Unicode, Bool
from traitlets.config import LoggingConfigurable

log = logging.getLogger('bookstore_config')


class BookstoreSettings(LoggingConfigurable):
    """Configuration for archival and publishing.

    Settings include storage directory locations, S3 authentication,
    additional S3 settings, and Bookstore resources.

    S3 authentication settings can be set, or they can be left unset when
    IAM is used.

    Like the Jupyter notebook, bookstore uses traitlets to handle
    configuration, loading from files or CLI.

    Attributes
    ----------
    workspace_prefix : str(``workspace``)
                       Directory to use for user workspace storage
    published_prefix : str(``published``)
                       Directory to use for published notebook storage
    s3_access_key_id : str, optional
                       Environment variable ``JPYNB_S3_ACCESS_KEY_ID``
    s3_secret_access_key : str, optional
                       Environment variable ``JPYNB_S3_SECRET_ACCESS_KEY``
    s3_endpoint_url : str(``"https://s3.amazonaws.com"``)
                       Environment variable ``JPYNB_S3_ENDPOINT_URL``
    s3_region_name : str(``"us-east-1"``)
                       Environment variable ``JPYNB_S3_REGION_NAME``
    s3_bucket : str(``""``)
                Bucket name, environment variable ``JPYNB_S3_BUCKET``
    max_threads : int(``16``)
                  Maximum threads from the threadpool available for S3 read/writes
    enable_s3_cloning : bool(``True``)
                        Enable cloning from s3.
    fs_cloning_basedir : str(``"/Users/jupyter"``)
                        Absolute path to base directory used to clone from the local file system
                  
    """

    workspace_prefix = Unicode("workspace", help="Prefix for the live workspace notebooks").tag(
        config=True
    )
    published_prefix = Unicode("published", help="Prefix for published notebooks").tag(config=True)
    enable_s3_cloning = Bool(True, help="Enable cloning from s3.").tag(config=True)

    s3_access_key_id = Unicode(
        help="S3/AWS access key ID", allow_none=True, default_value=None
    ).tag(config=True, env="JPYNB_S3_ACCESS_KEY_ID")
    s3_secret_access_key = Unicode(
        help="S3/AWS secret access key", allow_none=True, default_value=None
    ).tag(config=True, env="JPYNB_S3_SECRET_ACCESS_KEY")

    s3_endpoint_url = Unicode("https://s3.amazonaws.com", help="S3 endpoint URL").tag(
        config=True, env="JPYNB_S3_ENDPOINT_URL"
    )
    s3_region_name = Unicode("us-east-1", help="Region name").tag(
        config=True, env="JPYNB_S3_REGION_NAME"
    )
    s3_bucket = Unicode("", help="Bucket name to store notebooks").tag(
        config=True, env="JPYNB_S3_BUCKET"
    )

    max_threads = Integer(
        16, help="Maximum number of threads for the threadpool allocated for S3 read/writes"
    ).tag(config=True)

    fs_cloning_basedir = Unicode(
        "", help=("Absolute path to base directory used to clone from the local file system")
    ).tag(config=True)


def validate_bookstore(settings: BookstoreSettings):
    """Check that settings exist.

    Parameters
    ----------
    settings : bookstore.bookstore_config.BookstoreSettings
               Instantiated settings object to be validated.

    Returns
    -------
    validation_checks : dict
        Statements about whether features are validly configured and available
    """
    general_settings = [settings.s3_bucket != "", settings.s3_endpoint_url != ""]
    archive_settings = [*general_settings, settings.workspace_prefix != ""]
    published_settings = [*general_settings, settings.published_prefix != ""]
    s3_cloning_settings = [settings.enable_s3_cloning]
    fs_cloning_settings = [Path(settings.fs_cloning_basedir).is_absolute()]

    validation_checks = {
        "bookstore_valid": all(general_settings),
        "archive_valid": all(archive_settings),
        "publish_valid": all(published_settings),
        "s3_clone_valid": all(s3_cloning_settings),
        "fs_clone_valid": all(fs_cloning_settings),
    }
    if not validation_checks["fs_clone_valid"] and settings.fs_cloning_basedir != "":
        log.info(
            f"{settings.fs_cloning_basedir} is not an absolute path, file system cloning will be disabled."
        )

    return validation_checks
